@@ -1,8 +1,10 @@
-#!/usr/bin/python
+#!/usr/bin/python3
 
 import os
 import unittest
 
+import apt_pkg
+apt_pkg.config.set("Dir", "./aptroot")
 import apt
 
 import unattended_upgrade
@@ -16,7 +18,7 @@ class MockOptions(object):
     download_only = False
     dry_run = True
     apt_debug = False
-    minimal_upgrade_steps = False
+    minimal_upgrade_steps = True
 
 
 class TestRewindCache(unittest.TestCase):
@@ -38,8 +40,12 @@ class TestRewindCache(unittest.TestCase):
         to_upgrade, kept_back = unattended_upgrade.calculate_upgradable_pkgs(
             self.cache, options, self.allowed_origins, blacklisted_pkgs,
             whitelisted_pkgs)
-        self.assertEqual(to_upgrade, [self.cache["test-package"]])
+        self.assertEqual(to_upgrade, [self.cache[p] for p
+                                      in ["test-package", "test2-package",
+                                          "test3-package"]])
         self.assertEqual(kept_back, ["z-package"])
+        unattended_upgrade.rewind_cache(self.cache, to_upgrade)
+        self.assertEqual(self.cache['test-package'].candidate.version, "2.0")
 
 
 if __name__ == "__main__":
