@@ -39,15 +39,14 @@ import re
 import os
 import select
 import signal
+import socket
 import string
 import subprocess
 import sys
 
 try:
-    from typing import AbstractSet, DefaultDict, Dict, Iterable, List, Tuple
-    from typing import Union
+    from typing import AbstractSet, Dict, Iterable, List, Tuple, Union
     AbstractSet  # pyflakes
-    DefaultDict  # pyflakes
     Dict  # pyflakes
     Iterable  # pyflakes
     List  # pyflakes
@@ -56,7 +55,6 @@ try:
 except ImportError:
     pass
 
-from collections import defaultdict
 from datetime import date
 from email.message import Message
 from gettext import gettext as _
@@ -77,7 +75,6 @@ import apt_pkg
 
 # the reboot required flag file used by packages
 REBOOT_REQUIRED_FILE = "/var/run/reboot-required"
-KEPT_PACKAGES_FILE = "var/lib/unattended-upgrades/kept-back"
 MAIL_BINARY = "/usr/bin/mail"
 SENDMAIL_BINARY = "/usr/sbin/sendmail"
 USERS = "/usr/bin/users"
@@ -233,8 +230,8 @@ class UnattendedUpgradesCache(apt.Cache):
         for marked_pkg in self.get_changes():
             if marked_pkg.name in self._cached_candidate_pkgnames:
                 continue
-            if not is_in_allowed_origin(marked_pkg.candidate,
-                                        self.allowed_origins):
+            if not is_allowed_origin(marked_pkg.candidate,
+                                     self.allowed_origins):
                 try:
                     ver_in_allowed_origin(marked_pkg,
                                           self.allowed_origins)
@@ -415,36 +412,28 @@ class Unlocked:
             apt_pkg.pkgsystem_lock()
 
 
-class KeptPkgs(defaultdict):
+class UnattendedUpgradesResult:
     """
-    Packages to keep by highest allowed pretty-printed origin
-
+    Represent the (potentially partial) results of an unattended-upgrades
+    run
     """
-    def add(self, pkg, cache):
-        # type: (apt.Package, UnattendedUpgradesCache) -> None
-        allowed_origins = cache.allowed_origins
-        for origin in ver_in_allowed_origin(pkg, allowed_origins).origins:
-            if is_allowed_origin(origin, allowed_origins):
-                self[origin.origin + " " + origin.archive].add(pkg.name)
-                return
-
-    def pop_upgradable(self, cache):
-        # type: (UnattendedUpgradesCache) -> List[apt.Package]
-        upgradable = []
-        empty_sets = set()
-        for name, pkg_set in self.items():
-            remove_from_set = set()
-            for pkg_name in pkg_set:
-                pkg = cache[pkg_name]
-                if pkg.marked_install or pkg.marked_upgrade:
-                    remove_from_set.add(pkg_name)
-                    upgradable.append(pkg)
-            pkg_set -= remove_from_set
-            if not pkg_set:
-                empty_sets.add(name)
-        for empty_set in empty_sets:
-            del(self[empty_set])
-        return upgradable
+    def __init__(self,
+                 success,                 # type: bool
+                 result_str="",           # type: str
+                 pkgs=[],                 # type: List[str]
+                 pkgs_kept_back=[],       # type: List[str]
+                 pkgs_removed=[],         # type: List[str]
+                 pkgs_kept_installed=[],  # type: List[str]
+                 update_stamp=False       # type: bool
+                 ):
+        # type: (...) -> None
+        self.success = success
+        self.result_str = result_str
+        self.pkgs = pkgs
+        self.pkgs_kept_back = pkgs_kept_back
+        self.pkgs_removed = pkgs_removed
+        self.pkgs_kept_installed = pkgs_kept_installed
+        self.update_stamp = update_stamp
 
 
 def is_dpkg_journal_dirty():
@@ -567,7 +556,7 @@ def match_whitelist_string(whitelist, origin):
         return False
     res = True
     # make "\," the html quote equivalent
-    whitelist = whitelist.replace("\,", "%2C")
+    whitelist = whitelist.replace("\\,", "%2C")
     for token in whitelist.split(","):
         # strip and unquote the "," back
         (what, value) = [s.strip().replace("%2C", ",")
@@ -608,13 +597,6 @@ def cache_commit(cache,           # type: apt.Cache
                  ):
     # type: (...) -> Tuple[bool, Exception]
     """Commit the changes from the given cache to the system"""
-    # set debconf to NON_INTERACTIVE, redirect output
-    os.putenv("DEBIAN_FRONTEND", "noninteractive")
-
-    # only act if there is anything to do (important to avoid writing
-    # empty log stanzas)
-    if len(cache.get_changes()) == 0:
-        return True, None
 
     error = None
     res = False
@@ -632,7 +614,6 @@ def cache_commit(cache,           # type: apt.Cache
         error = e
         if verbose:
             logging.exception("Exception happened during upgrade.")
-            res = False
         cache.clear()
     return res, error
 
@@ -652,12 +633,12 @@ def upgrade_normal(cache, logfile_dpkg, verbose):
 
 def use_minimal_steps(options):
     # type: (Options) -> bool
-    return (options.minimal_upgrade_steps or
-            # COMPAT with the mispelling
-            (apt_pkg.config.find_b("Unattended-Upgrades::MinimalSteps",
-                                   True) and
-             apt_pkg.config.find_b("Unattended-Upgrade::MinimalSteps",
-                                   True)))
+    # COMPAT with the mispelling
+    return (options.minimal_upgrade_steps
+            or (apt_pkg.config.find_b("Unattended-Upgrades::MinimalSteps",
+                                      True)
+                and apt_pkg.config.find_b("Unattended-Upgrade::MinimalSteps",
+                                          True)))
 
 
 def upgrade_in_minimal_steps(cache,            # type: apt.Cache
@@ -688,7 +669,7 @@ def upgrade_in_minimal_steps(cache,            # type: apt.Cache
             pkg.mark_install()
         else:
             continue
-        upgrade_set_sizes[pkgname] = len(cache.get_changes())
+        upgrade_set_sizes[pkgname] = cache.install_count
     cache.clear()
 
     # to upgrade contains the package names
@@ -732,8 +713,9 @@ def upgrade_in_minimal_steps(cache,            # type: apt.Cache
 
         # write progress log information
         if len(pkgs_to_upgrade) > 0:
-            percent = ((len(pkgs_to_upgrade) - len(to_upgrade)) /
-                       float(len(pkgs_to_upgrade)) * 100.0)
+            all_count = len(pkgs_to_upgrade)
+            remaining_count = all_count - len(to_upgrade)
+            percent = remaining_count / float(all_count * 100.0)
         else:
             percent = 100.0
         install_log.status_change(pkg=",".join(changes),
@@ -759,21 +741,14 @@ def upgrade_in_minimal_steps(cache,            # type: apt.Cache
     return res
 
 
-def is_allowed_origin(origin, allowed_origins):
-    # type: (apt.package.Origin, List[str]) -> bool
-    for allowed in allowed_origins:
-        if match_whitelist_string(allowed, origin):
-            return True
-    return False
-
-
-def is_in_allowed_origin(ver, allowed_origins):
+def is_allowed_origin(ver, allowed_origins):
     # type: (apt.package.Version, List[str]) -> bool
     if not ver:
         return False
     for origin in ver.origins:
-        if is_allowed_origin(origin, allowed_origins):
-            return True
+        for allowed in allowed_origins:
+            if match_whitelist_string(allowed, origin):
+                return True
     return False
 
 
@@ -788,7 +763,7 @@ def ver_in_allowed_origin(pkg, allowed_origins):
                 continue
         except AttributeError:
             pass
-        if is_in_allowed_origin(ver, allowed_origins):
+        if is_allowed_origin(ver, allowed_origins):
             # leave as soon as we have the highest new candidate
             return ver
     raise NoAllowedOriginError()
@@ -869,7 +844,8 @@ def check_changes_for_sanity(cache, allowed_origins, blacklist, whitelist,
     # If there are no packages to be installed they were kept back
     if cache.install_count == 0:
         return False
-    for pkg in cache.get_changes():
+    changes = cache.get_changes()
+    for pkg in changes:
         if pkg.marked_delete:
             logging.debug("pkg %s now marked delete" % pkg.name)
             return False
@@ -879,7 +855,7 @@ def check_changes_for_sanity(cache, allowed_origins, blacklist, whitelist,
             if not any([o.trusted for o in pkg.candidate.origins]):
                 logging.debug("pkg %s is untrusted" % pkg.name)
                 return False
-            if not is_in_allowed_origin(pkg.candidate, allowed_origins):
+            if not is_allowed_origin(pkg.candidate, allowed_origins):
                 logging.debug("pkg %s not in allowed origin" % pkg.name)
                 return False
             if not is_pkg_change_allowed(pkg, blacklist, whitelist):
@@ -888,17 +864,25 @@ def check_changes_for_sanity(cache, allowed_origins, blacklist, whitelist,
             ignore_require_restart = apt_pkg.config.find_b(
                 "Unattended-Upgrade::IgnoreAppsRequireRestart", False)
             upgrade_requires = pkg.candidate.record.get("Upgrade-Requires")
-            if (pkg.marked_upgrade and ignore_require_restart is False and
-                    upgrade_requires == "app-restart"):
+            if pkg.marked_upgrade and ignore_require_restart is False \
+               and upgrade_requires == "app-restart":
                 logging.debug("pkg %s requires app-restart, not safe to "
                               "upgrade unattended")
                 return False
     # check that the package we want to upgrade is in the change set
-    if desired_pkg and desired_pkg not in cache.get_changes():
+    if desired_pkg and desired_pkg not in changes:
         return False
     return True
 
 
+def is_deb(file):
+    # type: (str) -> bool
+    if re.match("^.*\\.deb$", file):
+        return True
+    else:
+        return False
+
+
 def pkgname_from_deb(debfile):
     # type: (str) -> str
     # FIXME: add error checking here
@@ -999,7 +983,7 @@ def conffile_prompt(destFile, prefix=""):
         print(_("error message: %s") % e)
         logging.error(_("Apt returned an error, exiting"))
         logging.error(_("error message: %s"), e)
-        sys.exit(1)
+        raise
     except LookupError as e:
         logging.debug("No conffiles in deb %s (%s)" % (destFile, e))
     if not pkg_conffiles:
@@ -1098,7 +1082,7 @@ def rewind_cache(cache, pkgs_to_upgrade):
 
 def host():
     # type: () -> str
-    return os.uname()[1]
+    return socket.getfqdn()
 
 
 # *sigh* textwrap is nice, but it breaks "linux-image" into two
@@ -1167,9 +1151,10 @@ def _send_mail_using_sendmail(from_address, to_address, subject, body):
     return ret
 
 
-def send_summary_mail(pkgs,                 # type: str
+def send_summary_mail(pkgs,                 # type: List[str]
                       res,                  # type: bool
-                      pkgs_kept_back,       # type: KeptPkgs
+                      result_str,           # type: str
+                      pkgs_kept_back,       # type: List[str]
                       pkgs_removed,         # type: List[str]
                       pkgs_kept_installed,  # type: List[str]
                       mem_log,              # type: StringIO
@@ -1197,15 +1182,16 @@ def send_summary_mail(pkgs,                 # type: str
     reboot_flag_str = _(
         "[reboot required]") if os.path.isfile(REBOOT_REQUIRED_FILE) else ""
     # Check if packages are kept on hold
-    hold_flag_str = _("[package on hold]") if (pkgs_kept_back or
-                                               pkgs_kept_installed) else ""
+    hold_flag_str = (_("[package on hold]") if pkgs_kept_back
+                     or pkgs_kept_installed else "")
     logging.debug("Sending mail to %s" % to_email)
     subject = _(
         "{hold_flag}{reboot_flag} unattended-upgrades result for "
         "{machine}: {result}").format(
             hold_flag=hold_flag_str, reboot_flag=reboot_flag_str,
-            machine=host(), result=res).strip()
-    body = _("Unattended upgrade returned: %s\n\n") % res
+            machine=host(), result="SUCCESS" if res else "FAILURE").strip()
+    body = wrap(_("Unattended upgrade result: %s") % result_str, 70, " ")
+    body += "\n\n"
     if os.path.isfile(REBOOT_REQUIRED_FILE):
         body += _(
             "Warning: A reboot is required to complete this upgrade.\n\n")
@@ -1213,14 +1199,13 @@ def send_summary_mail(pkgs,                 # type: str
         body += _("Packages that were upgraded:\n")
     else:
         body += _("Packages that attempted to upgrade:\n")
-    body += " " + wrap(pkgs, 70, " ")
+    body += " " + wrap(" ".join(pkgs), 70, " ")
     body += "\n"
     if pkgs_kept_back:
         body += _("Packages with upgradable origin but kept back:\n")
-        for origin, origin_pkgs in pkgs_kept_back.items():
-            body += " " + origin + ":\n"
-            body += "  " + wrap(" ".join(origin_pkgs), 70, "  ") + "\n"
+        body += " " + wrap(" ".join(pkgs_kept_back), 70, " ")
         body += "\n"
+    body += "\n"
     if pkgs_removed:
         body += _("Packages that were auto-removed:\n")
         body += " " + wrap(" ".join(pkgs_removed), 70, " ")
@@ -1259,8 +1244,7 @@ def do_install(cache,             # type: apt.Cache
                logfile_dpkg,      # type: str
                ):
     # type: (...) -> bool
-    # set debconf to NON_INTERACTIVE, redirect output
-    os.putenv("DEBIAN_FRONTEND", "noninteractive")
+
     setup_apt_listchanges()
 
     logging.info(_("Writing dpkg log to %s"), logfile_dpkg)
@@ -1280,7 +1264,6 @@ def do_install(cache,             # type: apt.Cache
                 blacklisted_pkgs, whitelisted_pkgs, logfile_dpkg,
                 options.verbose or options.debug)
         else:
-            mark_pkgs_to_upgrade(cache, pkgs_to_upgrade)
             pkg_install_success = upgrade_normal(
                 cache, logfile_dpkg, options.verbose or options.debug)
     except Exception as e:
@@ -1440,7 +1423,7 @@ def write_stamp_file():
 
 def try_to_upgrade(pkg,               # type: apt.Package
                    pkgs_to_upgrade,   # type: List[apt.Package]
-                   pkgs_kept_back,    # type: KeptPkgs
+                   pkgs_kept_back,    # type: List[str]
                    cache,             # type: apt.Cache
                    allowed_origins,   # type: List[str]
                    blacklisted_pkgs,  # type: List[str]
@@ -1466,20 +1449,24 @@ def try_to_upgrade(pkg,               # type: apt.Package
             # re-eval pkgs_kept_back as the resolver may fail to
             # directly upgrade a pkg, but that may work during
             # a subsequent operation, see debian bug #639840
-            pkgs_to_upgrade.extend(pkgs_kept_back.pop_upgradable(cache))
+            for pkgname in pkgs_kept_back:
+                if cache[pkgname].marked_install \
+                   or cache[pkgname].marked_upgrade:
+                    pkgs_kept_back.remove(pkgname)
+                    pkgs_to_upgrade.append(cache[pkgname])
         else:
             logging.debug("sanity check failed for: %s"
                           % str({str(p.candidate)
                                  for p in cache.get_changes()}))
             rewind_cache(cache, pkgs_to_upgrade)
-            pkgs_kept_back.add(pkg, cache)
+            pkgs_kept_back.append(pkg.name)
     except (SystemError, NoAllowedOriginError) as e:
         # can't upgrade
         logging.warning(
             _("package %s upgradable but fails to "
                 "be marked for upgrade (%s)"), pkg.name, e)
         rewind_cache(cache, pkgs_to_upgrade)
-        pkgs_kept_back.add(pkg, cache)
+        pkgs_kept_back.append(pkg.name)
 
 
 def calculate_upgradable_pkgs(cache,             # type: apt.Cache
@@ -1488,9 +1475,9 @@ def calculate_upgradable_pkgs(cache,             # type: apt.Cache
                               blacklisted_pkgs,  # type: List[str]
                               whitelisted_pkgs,  # type: List[str]
                               ):
-    # type: (...) -> Tuple[List[apt.Package], KeptPkgs]
+    # type: (...) -> Tuple[List[apt.Package], List[str]]
     pkgs_to_upgrade = []  # type: List[apt.Package]
-    pkgs_kept_back = KeptPkgs(set)
+    pkgs_kept_back = []   # type: List[str]
 
     # now do the actual upgrade
     for pkg in cache:
@@ -1498,8 +1485,8 @@ def calculate_upgradable_pkgs(cache,             # type: apt.Cache
             logging.debug("Checking: %s (%s)" % (
                 pkg.name, getattr(pkg.candidate, "origins", [])))
 
-        if (pkg.is_upgradable and
-                is_pkgname_in_whitelist(pkg.name, whitelisted_pkgs)):
+        if (pkg.is_upgradable
+                and is_pkgname_in_whitelist(pkg.name, whitelisted_pkgs)):
             try:
                 ver_in_allowed_origin(pkg, allowed_origins)
             except NoAllowedOriginError:
@@ -1512,10 +1499,6 @@ def calculate_upgradable_pkgs(cache,             # type: apt.Cache
                            allowed_origins,
                            blacklisted_pkgs,
                            whitelisted_pkgs)
-
-    if cache.get_changes():
-        cache.clear()
-
     return pkgs_to_upgrade, pkgs_kept_back
 
 
@@ -1525,22 +1508,27 @@ def get_dpkg_log_content(logfile_dpkg, install_start_time):
         logfile_dpkg, install_start_time))
     content = []
     found_start = False
-    with io.open(logfile_dpkg, encoding='utf-8', errors='replace') as fp:
-        # read until we find the last "Log started: "
-        for line in fp.readlines():
-            # scan for the first entry we need (minimal-step mode
-            # creates a new stanza for each individual install)
-            if not found_start and line.startswith("Log started: "):
-                stanza_start = LoggingDateTime.from_string(
-                    line[len("Log started: "):-1])
-                if stanza_start >= install_start_time:
-                    found_start = True
-            if found_start:
-                # skip progress indicator until #860931 is fixed in apt + dpkg
-                if re.match("^\(Reading database \.\.\. ()|([0-9]+%)$", line):
-                    continue
-                content.append(line)
-    return "".join(content)
+    try:
+        with io.open(logfile_dpkg, encoding='utf-8', errors='replace') as fp:
+            # read until we find the last "Log started: "
+            for line in fp.readlines():
+                # scan for the first entry we need (minimal-step mode
+                # creates a new stanza for each individual install)
+                if not found_start and line.startswith("Log started: "):
+                    stanza_start = LoggingDateTime.from_string(
+                        line[len("Log started: "):-1])
+                    if stanza_start >= install_start_time:
+                        found_start = True
+                if found_start:
+                    # skip progress indicator until #860931 is fixed
+                    # in apt + dpkg
+                    if re.match("^\\(Reading database \\.\\.\\. ()|([0-9]+%)$",
+                                line):
+                        continue
+                    content.append(line)
+        return "".join(content)
+    except FileNotFoundError:
+        return ""
 
 
 def get_auto_removable(cache):
@@ -1696,51 +1684,86 @@ def is_update_day():
     return False
 
 
-def update_kept_packages(kept_pkgs, kept_file):
-    # type: (DefaultDict[str, List[str]], str) -> None
-    if kept_pkgs:
-        pkgs_all_origins = set()
-        for origin_pkgs in kept_pkgs.values():
-            pkgs_all_origins.update(origin_pkgs)
-        try:
-            with open(kept_file, "w") as kf:
-                kf.write(" ".join(sorted(pkgs_all_origins)))
-        except Exception:
-            logging.error(_("Could not open %s for saving list of packages "
-                            "kept back." % kept_file))
-    else:
-        if os.path.exists(kept_file):
-            os.remove(kept_file)
-
-
-def mark_pkgs_to_upgrade(cache, pkgs_to_upgrade):
-    # type (apt.Cache, List[apt.Package]) -> None
-    for pkg in pkgs_to_upgrade:
-        if pkg.is_upgradable:
-            cache.mark_upgrade_adjusted(pkg,
-                                        from_user=not pkg.is_auto_installed)
-        elif not pkg.is_installed:
-            cache.mark_install_adjusted(pkg, from_user=False)
-
-
-def main(options, rootdir="/"):
+def main(options, rootdir=""):
     # type: (Options, str) -> int
     # useful for testing
-    if not rootdir == "/":
+    if rootdir:
         _setup_alternative_rootdir(rootdir)
 
+    # see debian #776752
+    install_start_time = datetime.datetime.now().replace(microsecond=0)
+
     # setup logging
     mem_log = _setup_logging(options)
+    # get log
+    logfile_dpkg = os.path.join(_get_logdir(), 'unattended-upgrades-dpkg.log')
+    if not os.path.exists(logfile_dpkg):
+        with open(logfile_dpkg, 'w'):
+            pass
+
+    # lock for the shutdown check
+    shutdown_lock = apt_pkg.get_lock(LOCK_FILE)
+    if shutdown_lock < 0:
+        logging.error("Lock file is already taken, exiting")
+        return 1
+
+    try:
+        res = run(options, rootdir, mem_log, logfile_dpkg,
+                  install_start_time)
+        if res.result_str and not options.dry_run:
+            # there is some meaningful result which is worth an email
+            log_content = get_dpkg_log_content(logfile_dpkg,
+                                               install_start_time)
+            send_summary_mail(res.pkgs, res.success, res.result_str,
+                              res.pkgs_kept_back, res.pkgs_removed,
+                              res.pkgs_kept_installed, mem_log,
+                              log_content)
+        if res.update_stamp:
+            # write timestamp file
+            write_stamp_file()
+            if not options.dry_run:
+                # check if the user wants a reboot
+                reboot_if_requested_and_needed()
+        os.close(shutdown_lock)
+        if res.success:
+            return 0
+        else:
+            return 1
+
+    except Exception as e:
+        logger = logging.getLogger()
+        logger.exception(_("An error occurred: %s"), e)
+        log_content = get_dpkg_log_content(logfile_dpkg,
+                                           install_start_time)
+        if not options.dry_run:
+            send_summary_mail(["<unknown>"], False, _("An error occurred"), [],
+                              [], [],
+                              mem_log, log_content)
+        # Re-raise exceptions for apport
+        raise
+
+
+def run(options,             # type: Options
+        rootdir,             # type: str
+        mem_log,             # type: StringIO
+        logfile_dpkg,        # type: str
+        install_start_time,  # type: datetime.datetime
+        ):
+    # type: (...) -> UnattendedUpgradesResult
 
     # check if today is a patch day
     if not is_update_day():
-        return 0
+        return UnattendedUpgradesResult(True)
+
+    # check if u-u should be stopped already
+    if should_stop():
+        return UnattendedUpgradesResult(False)
 
     # check to see if want to auto-upgrade the devel release
     if "(development branch)" in DISTRO_DESC and not\
             apt_pkg.config.find_b("Unattended-Upgrade::DevRelease", False):
         logging.info(_("Not running on the development release."))
-        return 0
+        return UnattendedUpgradesResult(True)
     # format (origin, archive), e.g. ("Ubuntu","dapper-security")
     allowed_origins = get_allowed_origins()
 
@@ -1756,18 +1779,9 @@ def main(options, rootdir="/"):
 
     logging.info(_("Starting unattended upgrades script"))
 
-    # lock for the shutdown check
-    shutdown_lock = apt_pkg.get_lock(LOCK_FILE)
-    if shutdown_lock < 0:
-        logging.error("Lock file is already taken, exiting")
-        sys.exit(1)
-
     # display available origin
     logging.info(_("Allowed origins are: %s"), ", ".join(allowed_origins))
 
-    # see debian #776752
-    install_start_time = datetime.datetime.now().replace(microsecond=0)
-
     # check and get lock
     try:
         apt_pkg.pkgsystem_lock()
@@ -1775,7 +1789,8 @@ def main(options, rootdir="/"):
         logging.error(_("Lock could not be acquired (another package "
                         "manager running?)"))
         print(_("Cache lock can not be acquired, exiting"))
-        sys.exit(1)
+        return UnattendedUpgradesResult(
+            False, _("Lock could not be acquired"))
 
     # check if the journal is dirty and if so, take emergceny action
     # the alternative is to leave the system potentially unsecure until
@@ -1787,7 +1802,6 @@ def main(options, rootdir="/"):
             _("Unclean dpkg state detected, trying to correct"))
         print(_("Unclean dpkg state detected, trying to correct"))
         env = copy.copy(os.environ)
-        env["DEBIAN_FRONTEND"] = "noninteractive"
         env["DPKG_FRONTEND_LOCKED"] = "1"
         try:
             with Unlocked():
@@ -1808,12 +1822,14 @@ def main(options, rootdir="/"):
         print(_("error message: %s") % error)
         logging.error(_("Apt returned an error, exiting"))
         logging.error(_("error message: %s"), error)
-        sys.exit(1)
+        return UnattendedUpgradesResult(
+            False, _("Apt returned an error, exiting"))
 
     if cache._depcache.broken_count > 0:
         print(_("Cache has broken packages, exiting"))
         logging.error(_("Cache has broken packages, exiting"))
-        sys.exit(1)
+        return UnattendedUpgradesResult(
+            False, _("Cache has broken packages, exiting"))
 
     # FIXME: make this into a ContextManager
     # be nice when calculating the upgrade as its pretty CPU intensive
@@ -1834,16 +1850,15 @@ def main(options, rootdir="/"):
     pkgs_to_upgrade, pkgs_kept_back = calculate_upgradable_pkgs(
         cache, options, allowed_origins, blacklisted_pkgs, whitelisted_pkgs)
     pkgs_to_upgrade.sort(key=lambda p: p.name)
-    pkgs = "\n".join([pkg.name for pkg in pkgs_to_upgrade])
-    logging.debug("pkgs that look like they should be upgraded: %s" % pkgs)
+    pkgs = [pkg.name for pkg in pkgs_to_upgrade]
+    logging.debug("pkgs that look like they should be upgraded: %s"
+                  % "\n".join(pkgs))
 
     # FIXME: make this into a ContextManager
     # stop being nice
     os.nice(old_priority - os.nice(0))
 
     # download what looks good
-    mark_pkgs_to_upgrade(cache, pkgs_to_upgrade)
-
     if options.debug:
         fetcher = apt_pkg.Acquire(apt.progress.text.AcquireProgress())
     else:
@@ -1856,7 +1871,7 @@ def main(options, rootdir="/"):
     # TODO: download files one by one and check for stop request after each of
     # them
     if should_stop():
-        return 1
+        return UnattendedUpgradesResult(False, _("Upgrade was interrupted"))
     try:
         pm.get_archives(fetcher, list, recs)
     except SystemError as e:
@@ -1868,10 +1883,7 @@ def main(options, rootdir="/"):
         logging.error("fetch.run() result: %s", e)
 
     if options.download_only:
-        return 0
-
-    if cache.get_changes():
-        cache.clear()
+        return UnattendedUpgradesResult(True)
 
     pkg_conffile_prompt = False
     if dpkg_conffile_prompt():
@@ -1887,15 +1899,22 @@ def main(options, rootdir="/"):
                       item.desc_uri)
                 logging.error(_("The URI %s failed to download, aborting"),
                               item.desc_uri)
-                sys.exit(1)
+                return UnattendedUpgradesResult(
+                    False, (_("The URI %s failed to download, aborting") %
+                            item.desc_uri))
             if not os.path.exists(item.destfile):
                 print(_("Download finished, but file %s not there?!?") %
                       item.destfile)
                 logging.error("Download finished, but file %s not "
                               "there?!?", item.destfile)
-                sys.exit(1)
+                return UnattendedUpgradesResult(
+                    False, (_("Download finished, but file %s not there?!?") %
+                            item.destfile))
             if not item.is_trusted:
                 blacklisted_pkgs.append(pkgname_from_deb(item.destfile))
+            if not is_deb(item.destfile):
+                logging.debug("%s is not a .deb file" % item)
+                continue
             if conffile_prompt(item.destfile):
                 # skip package (means to re-run the whole marking again
                 # and making sure that the package will not be pulled in by
@@ -1912,10 +1931,8 @@ def main(options, rootdir="/"):
                 logging.warning(_("Package %s has conffile prompt and "
                                   "needs to be upgraded manually"),
                                 pkgname_from_deb(item.destfile))
-                pkg_name = pkgname_from_deb(item.destfile)
-                if not is_pkgname_in_blacklist(pkg_name, blacklisted_pkgs):
-                    blacklisted_pkgs.append("%s$" % re.escape(pkg_name))
-                pkgs_kept_back.add(cache[pkg_name], cache)
+                blacklisted_pkgs.append(pkgname_from_deb(item.destfile))
+                pkgs_kept_back.append(pkgname_from_deb(item.destfile))
                 pkg_conffile_prompt = True
 
         # redo the selection about the packages to upgrade based on the new
@@ -1925,6 +1942,7 @@ def main(options, rootdir="/"):
         logging.debug("whitelist: %s" % whitelisted_pkgs)
         # find out about the packages that are upgradable (in a allowed_origin)
         if len(blacklisted_pkgs) > 0 or len(whitelisted_pkgs) > 0:
+            cache.clear()
             old_pkgs_to_upgrade = pkgs_to_upgrade[:]
             pkgs_to_upgrade = []
             for pkg in old_pkgs_to_upgrade:
@@ -1939,25 +1957,16 @@ def main(options, rootdir="/"):
                     pkgs_to_upgrade.append(pkg)
                 else:
                     if not (pkg.name in pkgs_kept_back):
-                        pkgs_kept_back.add(pkg, cache)
+                        pkgs_kept_back.append(pkg.name)
                     logging.info(_("package %s not upgraded"), pkg.name)
                     cache.clear()
                     for pkg2 in pkgs_to_upgrade:
                         cache.call_adjusted(
                             apt.package.Package.mark_upgrade, pkg2,
                             from_user=not pkg2.is_auto_installed)
-            if cache.get_changes():
-                cache.clear()
-
     else:
         logging.debug("dpkg is configured not to cause conffile prompts")
 
-    # get log
-    logfile_dpkg = os.path.join(_get_logdir(), 'unattended-upgrades-dpkg.log')
-    if not os.path.exists(logfile_dpkg):
-        with open(logfile_dpkg, 'w'):
-            pass
-
     # auto-removals
     kernel_pkgs_remove_success = True  # type: bool
     kernel_pkgs_removed = []           # type: List[str]
@@ -1972,9 +1981,9 @@ def main(options, rootdir="/"):
         # of kernel packages
         auto_removable_kernel_pkgs = {
             p for p in auto_removable
-            if (cache.versioned_kernel_pkgs_regexp and
-                cache.versioned_kernel_pkgs_regexp.match(p) and
-                not cache.running_kernel_pkgs_regexp.match(p))}
+            if (cache.versioned_kernel_pkgs_regexp
+                and cache.versioned_kernel_pkgs_regexp.match(p)
+                and not cache.running_kernel_pkgs_regexp.match(p))}
         if auto_removable_kernel_pkgs:
             logging.info(_("Removing unused kernel packages: %s"),
                          " ".join(auto_removable_kernel_pkgs))
@@ -1995,31 +2004,28 @@ def main(options, rootdir="/"):
         pending_autoremovals = set()
 
     # exit if there is nothing to do and nothing to report
-    if (len(pending_autoremovals) == 0 and
-            len(pkgs_to_upgrade) == 0 and
-            len(pkgs_kept_back) == 0):
+    if (len(pending_autoremovals) == 0
+            and len(pkgs_to_upgrade) == 0
+            and len(pkgs_kept_back) == 0):
         logging.info(_("No packages found that can be upgraded unattended "
                        "and no pending auto-removals"))
-        # FIXME: DRY violation, write_stamp_file() is used below as well
-        write_stamp_file()
-        # check if we couldn't reboot on previous run because
-        # a user was logged-in at this time
-        os.close(shutdown_lock)
-        # never reboot during a dry run
-        if not options.dry_run:
-            reboot_if_requested_and_needed()
-        return 0
+        return UnattendedUpgradesResult(
+            kernel_pkgs_remove_success,
+            _("No packages found that can be upgraded unattended and no "
+              "pending auto-removals"),
+            pkgs_removed=kernel_pkgs_removed,
+            pkgs_kept_installed=kernel_pkgs_kept_installed,
+            update_stamp=True)
 
     # check if its configured for install on shutdown, if so, the
     # environment UNATTENDED_UPGRADES_FORCE_INSTALL_ON_SHUTDOWN will
     # be set by the unatteded-upgrades-shutdown script
-    if ("UNATTENDED_UPGRADES_FORCE_INSTALL_ON_SHUTDOWN" not in os.environ and
-            apt_pkg.config.find_b(
-                "Unattended-Upgrade::InstallOnShutdown", False)):
+    if ("UNATTENDED_UPGRADES_FORCE_INSTALL_ON_SHUTDOWN" not in os.environ
+        and apt_pkg.config.find_b(
+            "Unattended-Upgrade::InstallOnShutdown", False)):
         logger = logging.getLogger()
         logger.debug("Configured to install on shutdown, so exiting now")
-        os.close(shutdown_lock)
-        return 0
+        return UnattendedUpgradesResult(True)
 
     # check if we are in dry-run mode
     if options.dry_run:
@@ -2027,8 +2033,8 @@ def main(options, rootdir="/"):
         apt_pkg.config.set("Debug::pkgDPkgPM", "1")
 
     # do the install based on the new list of pkgs
-    pkgs = " ".join([pkg.name for pkg in pkgs_to_upgrade])
-    logging.info(_("Packages that will be upgraded: %s"), pkgs)
+    pkgs = [pkg.name for pkg in pkgs_to_upgrade]
+    logging.info(_("Packages that will be upgraded: %s"), " ".join(pkgs))
 
     # only perform install step if we actually have packages to install
     pkg_install_success = True
@@ -2042,14 +2048,15 @@ def main(options, rootdir="/"):
                                          logfile_dpkg)
     # Was the overall run succesful: only if everything installed
     # fine and nothing was held back because of a conffile prompt.
-    successful_run = (kernel_pkgs_remove_success and pkg_install_success and
-                      not pkg_conffile_prompt)
+    successful_run = (kernel_pkgs_remove_success and pkg_install_success
+                      and not pkg_conffile_prompt)
 
     # now check if any auto-removing needs to be done
     if cache._depcache.broken_count > 0:
         print(_("Cache has broken packages, exiting"))
         logging.error(_("Cache has broken packages, exiting"))
-        sys.exit(1)
+        return UnattendedUpgradesResult(
+            False, _("Cache has broken packages, exiting"), pkgs=pkgs)
 
     # make sure we start autoremovals with a clear cache
     if cache.get_changes():
@@ -2060,8 +2067,8 @@ def main(options, rootdir="/"):
     pkgs_removed = []         # type: List[str]
     pkgs_kept_installed = []  # type: List[str]
     if ((apt_pkg.config.find_b(
-            "Unattended-Upgrade::Remove-Unused-Dependencies", False) and
-         not SIGNAL_STOP_REQUEST)):
+            "Unattended-Upgrade::Remove-Unused-Dependencies", False)
+         and not SIGNAL_STOP_REQUEST)):
         auto_removals = get_auto_removable(cache)
         (pkg_remove_success,
          pkgs_removed,
@@ -2091,37 +2098,19 @@ def main(options, rootdir="/"):
                      cache._depcache.del_count,
                      cache._depcache.broken_count))
 
-    # send a mail (if needed)
-    if not options.dry_run:
-        log_content = get_dpkg_log_content(logfile_dpkg, install_start_time)
-        send_summary_mail(
-            pkgs, successful_run,
-            pkgs_kept_back,
-            kernel_pkgs_removed + pkgs_removed,
-            kernel_pkgs_kept_installed + pkgs_kept_installed,
-            mem_log, log_content)
-
     # clean after success install (if needed)
     keep_key = "Unattended-Upgrade::Keep-Debs-After-Install"
-    if (not apt_pkg.config.find_b(keep_key, False) and
-            not options.dry_run and
-            pkg_install_success):
+    if (not apt_pkg.config.find_b(keep_key, False)
+            and not options.dry_run
+            and pkg_install_success):
         clean_downloaded_packages(fetcher)
 
-    # FIXME: DRY violation, write_stamp_file() is used above as well
-    # write timestamp file
-    write_stamp_file()
-
-    os.close(shutdown_lock)
-    # check if the user wants a reboot
-    if not options.dry_run:
-        reboot_if_requested_and_needed()
-    if successful_run:
-        update_kept_packages(pkgs_kept_back,
-                             os.path.join(rootdir, KEPT_PACKAGES_FILE))
-        return 0
-    else:
-        return 1
+    return UnattendedUpgradesResult(
+        successful_run, _("All upgrades installed"), pkgs,
+        pkgs_kept_back,
+        kernel_pkgs_removed + pkgs_removed,
+        kernel_pkgs_kept_installed + pkgs_kept_installed,
+        update_stamp=True)
 
 
 class Options:
@@ -2140,6 +2129,9 @@ if __name__ == "__main__":
     gettext.bindtextdomain(localesApp, localesDir)
     gettext.textdomain(localesApp)
 
+    # set debconf to NON_INTERACTIVE
+    os.environ["DEBIAN_FRONTEND"] = "noninteractive"
+
     # this ensures the commandline is logged in /var/log/apt/history.log
     apt_pkg.config.set("Commandline::AsString", " ".join(sys.argv))
 
